# © Copyright Databand.ai, an IBM Company 2022

#### DOC START
from dbnd import output, parameter
from dbnd_spark import SparkTask


class PrepareData(SparkTask):
    data = parameter.data
    counters = output

    spark_conf_extension = {"spark.executor.memory": "1g"}


#### DOC END



# © Copyright Databand.ai, an IBM Company 2022

from dbnd import PythonTask, output, parameter, task


class TestDocRunATaskInPython:
    def test_doc(self):
        #### DOC START
        @task
        def calculate_alpha(alpha=0.5):
            return alpha

        class CalculateBeta(PythonTask):
            beta = parameter.value(0.1)
            result = output

            def run(self):
                self.result.write(self.beta)

        # if __name__ == "__main__":
        calculate_alpha.dbnd_run(alpha=0.4)
        CalculateBeta(beta=0.15).dbnd_run()
        #### DOC END



# © Copyright Databand.ai, an IBM Company 2022

import pyspark.sql as spark

from dbnd_examples_orchestration.dbnd_spark import spark_folder
from dbnd_examples_orchestration.orchestration.dbnd_spark.scripts import spark_script

from databand import output, parameter, pipeline
from databand.tasks import PipelineTask
from dbnd import log_dataframe, log_metric
from dbnd_spark.spark import PySparkInlineTask, PySparkTask, SparkTask, spark_task
from dbnd_spark.spark_config import SparkConfig
from targets import Target


COUNT_WITH_HTML_MAIN_CLASS = "ai.databand.examples.WordCountWithHtml"
WORD_COUNT_MAIN_CLASS = "ai.databand.examples.WordCount"


def _mvn_target_file(*path):
    return spark_folder("spark_jvm/target", *path)


class WordCountTask(SparkTask):
    text = parameter.data
    counters = output

    main_class = WORD_COUNT_MAIN_CLASS

    defaults = {SparkConfig.driver_memory: "2G"}

    def application_args(self):
        return [self.text, self.counters]


class WordCountPySparkTask(PySparkTask):
    text = parameter.data
    counters = output

    python_script = spark_script("word_count.py")

    def application_args(self):
        return [self.text, self.counters]


class WordCountPipeline(PipelineTask):
    text = parameter.data

    with_spark = output
    with_pyspark = output

    def band(self):
        self.with_spark = WordCountTask(text=self.text)
        self.with_pyspark = WordCountPySparkTask(text=self.text)


@pipeline
def word_count_new_cluster():
    wc = WordCountTask()

    from dbnd_gcp.dataproc.dataproc import DataProcCtrl

    create = DataProcCtrl(wc).create_engine()
    wc.set_upstream(create)


@spark_task(result=output[spark.DataFrame])
def word_count_inline(text=parameter.csv[spark.DataFrame], counters=output.txt.data):
    # type:  (spark.DataFrame, Target) -> spark.DataFrame
    from operator import add

    from dbnd_spark.spark import get_spark_session

    lines = text.rdd.map(lambda r: r[0])
    counts = (
        lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
    )
    counts.saveAsTextFile(str(counters))
    output = counts.collect()
    for (word, count) in output:
        print("%s: %i" % (word, count))

    counts_df = get_spark_session().createDataFrame(counts)
    log_dataframe("counts_df", counts_df)
    log_metric("test", 1)

    return counts_df


class WordCountSparkInline(PySparkInlineTask):
    text = parameter.csv[spark.DataFrame]
    counters = output.txt.data
    counters_auto_save = output[spark.DataFrame]

    def run(self):
        from operator import add

        from dbnd_spark.spark import get_spark_session

        lines = self.text.rdd.map(lambda r: r[0])
        counts = (
            lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
        )
        counts.saveAsTextFile(str(self.counters))
        output = counts.collect()
        for (word, count) in output:
            print("%s: %i" % (word, count))

        self.counters_auto_save = get_spark_session().createDataFrame(counts)



# © Copyright Databand.ai, an IBM Company 2022

"""
We want to define new custom object so we can run code like :
  dbnd run my_task --custom "some custom text for MyCustomObject constuctor"

"""

import logging

from typing import Dict, List, Set

from dbnd import PythonTask, output, parameter
from targets.values import ValueType, register_value_type


logger = logging.getLogger(__name__)


class MyCustomObject(object):
    def __init__(self, custom):
        self.custom = custom


class _MyCustomObjectValueType(ValueType):
    type = MyCustomObject

    def parse_from_str(self, x):
        return MyCustomObject(x)

    def to_str(self, x):
        return x.custom


register_value_type(_MyCustomObjectValueType())


class TaskWithCustomValue(PythonTask):
    custom_value = parameter.type(_MyCustomObjectValueType)
    list_of_customs = parameter.sub_type(_MyCustomObjectValueType)[List]

    report = output[str]

    def run(self):
        assert isinstance(self.custom_value, MyCustomObject)
        assert isinstance(self.list_of_customs[1], MyCustomObject)
        self.report = self.custom_value.custom + self.list_of_customs[1].custom


# We register this type, so now we can just use the .type -> MyCustomObject

register_value_type(_MyCustomObjectValueType())


class TaskWithCustomValueInline(PythonTask):
    # works only after registration!
    custom_value = parameter[MyCustomObject]
    custom_value_with_default = parameter.value(MyCustomObject("1"))

    list_of_customs = parameter.sub_type(MyCustomObject)[List]
    set_of_customs = parameter.sub_type(MyCustomObject)[Set]
    dict_of_customs = parameter.sub_type(MyCustomObject)[Dict]

    report = output[str]

    def run(self):
        assert isinstance(self.custom_value, MyCustomObject)
        assert isinstance(self.custom_value_with_default, MyCustomObject)

        self.report = "_".join(
            [
                c.custom
                for c in [
                    self.custom_value,
                    self.custom_value_with_default,
                    self.list_of_customs[0],
                    self.dict_of_customs["a"],
                ]
            ]
        )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import List

import dbnd

from dbnd import PythonTask, output


logger = logging.getLogger(__name__)


class PartitionedDataTask(PythonTask):
    """Create partitioned data"""

    partitioned_data = output.folder.csv.target

    def run(self):
        for i in range(10):
            self.partitioned_data.partition("part-%04d" % i).write(str(i))
        self.partitioned_data.mark_success()


class PartitionedDataReader(PythonTask):
    """Read partitioned data"""

    partitioned_data = dbnd.data[List[str]]
    concat = output.data

    def run(self):
        data = self.partitioned_data
        logger.info("Partitioned data read: %s", data)
        self.concat.write("".join(data))


class ExamplePartitionedDataPipeline(dbnd.PipelineTask):
    """Entry point of partitioned data pipeline"""

    concat = output.data

    def band(self):
        # This is a way to override every output of underneath tasks with custom output location (see "_custom")
        with dbnd.dbnd_config(
            config_values={
                "task": {
                    "task_output_path_format": "{root}/{env_label}/{task_family}{task_class_version}_custom/"
                    "{output_name}{output_ext}/date={task_target_date}"
                }
            }
        ):

            partitioned_data = PartitionedDataTask().partitioned_data
            self.concat = PartitionedDataReader(
                partitioned_data=partitioned_data
            ).concat



# © Copyright Databand.ai, an IBM Company 2022

import logging

import pandas as pd

from dbnd import PipelineTask, PythonTask, output, parameter


logger = logging.getLogger(__name__)


class PStepA(PythonTask):
    """First step of partitioned calculations"""

    data_partitioned = output.folder.csv.data

    # noinspection PyMethodMayBeStatic
    def _yield_data(self):
        for i in range(10):
            data = [["some_value", i]]
            yield pd.DataFrame(data=data, columns=["Names", "Value"])

    def run(self):
        self.data_partitioned.write_df(self._yield_data())


# noinspection PyMethodMayBeStatic
class PStepB(PythonTask):
    """Second step of partitioned calculations"""

    data_partitioned = parameter.data
    step_b_value = output.folder.csv.data

    def _stream_function(self, data_gen):
        for partition in data_gen:
            partition["StepB"] = partition["Value"] * 3
            yield partition

    def run(self):
        data = self.data_partitioned.read_df_partitioned()
        self.step_b_value.write_df(self._stream_function(data))


class ExamplePartitionedCalculations(PipelineTask):
    """Entry point of partitioned calculations"""

    step_b_value = output.data

    def band(self):
        data_partitioned = PStepA().data_partitioned
        self.step_b_value = PStepB(data_partitioned=data_partitioned).step_b_value



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.data import data_repo
from dbnd_examples_orchestration.orchestration.examples.salad import prepare_salad
from pytest import fixture

from dbnd_run.testing.helpers import assert_run_task
from targets import target


class TestSaladExample(object):
    def test_salad_deco_local__prepare(self):
        salad = prepare_salad.t()
        assert_run_task(salad)

    def test_prepare_salad_deco(self):
        assert_run_task(prepare_salad.t(vegetables=data_repo.vegetables))

    @fixture
    def vegetables(self, tmpdir):
        t = target(str(tmpdir), "vegetables.csv")
        t.write("a\nb\n")
        return t

    def test_regular_invoke(self, vegetables):
        salad = prepare_salad(vegetables.readlines(), "mayo")
        assert len(salad) == 2

    def test_as_task(self, vegetables):
        assert_run_task(prepare_salad.t(vegetables=vegetables))



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import os

from dbnd._core.utils.project.project_fs import abs_join
from dbnd.utils import timezone


def _dbnd_examples_project_root():
    # if env var exists - use it as the examples dir, otherwise, calculate relative from here.
    dbnd_examples_project = os.getenv("DBND_EXAMPLES_PATH", None)
    if dbnd_examples_project:
        return dbnd_examples_project

    return abs_join(__file__, "..", "..", "..")


def dbnd_examples_project_path(*path):
    return os.path.join(_dbnd_examples_project_root(), *path)


def dbnd_examples_src_path(*path):
    return dbnd_examples_project_path("src", "dbnd_examples_orchestration", *path)


def dbnd_examples_data_path(*path):
    return dbnd_examples_project_path("data", *path)


class ExamplesData(object):
    vegetables = dbnd_examples_data_path("vegetables_for_greek_salad.txt")
    wines = dbnd_examples_data_path("wine_quality_minimized.csv")
    wines_per_date = dbnd_examples_data_path("wine_quality_per_day")

    wines_full = dbnd_examples_data_path("wine_quality.csv.gz")

    raw_logs = timezone.parse("2018-06-22")
    partitioned_data_target_date = datetime.date(year=2018, month=9, day=3)
    partitioned_data = dbnd_examples_data_path("partitioned_data")


data_repo = ExamplesData()


def partner_data_file(*path):
    return dbnd_examples_data_path("example_raw_data", *path)


class PartnerData(object):
    seed = dbnd_examples_data_path("pipeline_train_seed_data.csv")

    def partner_a_file(self, date):
        return partner_data_file("partner_a", date.strftime("%Y-%m-%d") + ".csv")

    def partner_b_file(self, date):
        return partner_data_file("partner_b", date.strftime("%Y-%m-%d") + ".json")

    def partner_c_file(self, date):
        return partner_data_file("partner_c", date.strftime("%Y-%m-%d") + ".csv")

    def partner_to_file(self, name, date):
        return {
            "a": self.partner_a_file(date),
            "b": self.partner_b_file(date),
            "c": self.partner_c_file(date),
        }[name]

    def sample_customer_file(self):
        return partner_data_file("customer_a.csv")


demo_data_repo = PartnerData()



# © Copyright Databand.ai, an IBM Company 2022

from pytest import raises

from dbnd import PythonTask, parameter


class PrepareData(PythonTask):
    data = parameter.value("default")

    def run(self):
        return self.data


class TestDocAccessingTaskMeta:
    def test_doc(self):
        #### DOC START
        prepared_data = PrepareData(data="just some data")
        print(prepared_data)  # --> "prepare_data__19d1172be0"

        print(prepared_data.task_namespace)  # --> ""
        print(prepared_data.task_family)  # --> "prepare_data"
        print(prepared_data.task_id)  # --> "prepare_data__19d1172be0"

        print(PrepareData.task_namespace)  # --> ""
        print(PrepareData.task_family)  # --> "<property object at 0x1111e9548>"
        with raises(AttributeError):
            PrepareData.task_id  # --> Error!
        #### DOC END

    def test_current_task_object(self):
        #### DOC START
        from dbnd import current_task, task

        @task
        def calculate_alpha(alpha: int = 0.5):
            return current_task().task_version

        #### DOC END
        calculate_alpha.dbnd_run()

    def test_current_env_inside_task_or_pipeline(self):
        #### DOC START
        from dbnd import current_task, task

        @task
        def calculate_alpha(alpha: int = 0.5):
            return current_task().task_env.name  # The environment of the task
            # See EnvConfig object for all properties

        #### DOC END
        calculate_alpha.dbnd_run()



# © Copyright Databand.ai, an IBM Company 2022

import itertools
import logging

import tensorflow as tf

from dbnd_examples_orchestration.data import data_repo

from databand.parameters import TargetParameter
from dbnd import PipelineTask, PythonTask, data, output, parameter


logger = logging.getLogger(__name__)


# Dummy Class to return
class Dummy:
    pass


# language=bash
example1 = """
# run this task
dbnd run -m dbnd_examples_orchestration.tf_examples DataStats --DataStats-raw-data data/wine_quality.csv
"""


class DataStats(PythonTask):
    raw_data = data.target
    statistics = output.json

    def _validate_input(self, stat):
        # sample check, add your logic here
        if stat.iloc[0].max() != 4898:
            raise Exception("Failed to validate")

    def run(self):
        df = self.raw_data.read_df()
        #  this is just an example, your logic goes here
        stat = df.describe()
        self._validate_input(stat)
        stat.to_target(self.statistics)


# language=bash
example2 = """
# run this task separetly
dbnd run -m dbnd_examples_orchestration.example_tensorflow TransformToTfRecords --TransformToTfRecords-raw-data data/wine_quality.csv
"""


class TransformToTfRecords(PythonTask):
    raw_data = data.target
    tf_records = output

    def run(self):
        # this is just an example, your transformation goes here
        input_data = self.raw_data.read_df().values
        with tf.io.TFRecordWriter(self.tf_records.path) as writer:
            for row in input_data:
                example = tf.train.Example(
                    features=tf.train.Features(
                        feature={
                            "x": tf.train.Feature(
                                float_list=tf.train.FloatList(value=row)
                            )
                        }
                    )
                )
                writer.write(example.SerializeToString())


# This class is just a placeholder to configure one instance of model training. For now, it just manage params
# per what we discussed.
class TrainTfModel(PythonTask):
    data = data

    block1 = parameter(description="class name for block 1")[str]
    block2 = parameter(description="class name for block 2")[str]
    block3 = parameter(description="class name for block 3")[str]

    precision = parameter[int]
    model = output

    def run(self):
        # Example: get blocks of tensorflow plan by class name
        b1 = self._get_block_by_name(self.block1)
        b2 = self._get_block_by_name(self.block2)
        b3 = self._get_block_by_name(self.block3)
        logger.info("b1=%s b2=%s b3=%s", b1, b2, b3)
        # TODO here tensorflow graph should be constructed and executed. The output would be a model, saved as a file,
        # next line is just a demo
        self.model.write_pickle(self)

    def _get_block_by_name(self, class_name):
        # module = importlib.import_module(self.module_name)
        # class_ = getattr(module, class_name)
        # for now, return dummy class
        class_ = globals()["Dummy"]
        return class_()


class GenerateReport(PythonTask):
    model = data
    report = output

    def run(self):
        # for now, just a stub
        logger.log("This is a report for %s" % self.model)

        self.report.write("Some report!")


# this is to show how you can program multiple experiments and run them
# in the future you would be able to run them in parallel
class Experiment(PipelineTask):
    data = data(data_repo.wines)
    model = output
    validation_stats = output

    block1 = parameter.value(default="a")
    block2 = parameter.value(default="b")
    block3 = parameter.value(default="c")

    precision = parameter[int].default(32)

    def band(self):
        self.validation_stats = DataStats(raw_data=self.data).statistics
        transformed_data = TransformToTfRecords(raw_data=self.data)
        # transformed_data.set_upstream(validation_stats)

        self.model = TrainTfModel(
            data=transformed_data.tf_records,
            block1=self.block1,
            block2=self.block2,
            block3=self.block3,
            precision=self.precision,
        ).model


class ParameterSearch(PipelineTask):
    data = TargetParameter.data(default=data_repo.wines)
    model = output

    block1_list = ["a1", "a2", "a3"]
    block2_list = ["b1", "b2", "b3"]
    block3_list = ["c1", "c2", "c3"]

    def band(self):
        models = []
        for a, b, c in itertools.product(
            self.block1_list, self.block2_list, self.block3_list
        ):
            exp = Experiment(data=self.data, block1=a, block2=b, block3=c, precision=64)
            models[exp.task_id] = exp
        self.model = models



# © Copyright Databand.ai, an IBM Company 2022


from datetime import timedelta

from airflow import DAG
from airflow.utils.dates import days_ago

from dbnd import parameter, pipeline


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": days_ago(2),
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
}


class TestDocDecoratedDags:
    def test_prepare_data_jinja_templating(self):
        #### DOC START
        @pipeline
        def current_date(p_date=None):
            return p_date

        with DAG(dag_id=f"current_date_dag", default_args=default_args) as dag:
            current_date(p_date="{{ ts }}")
        #### DOC END

    def test_prepare_data_no_jinja_templating(self):
        #### DOC START
        @pipeline
        def current_date(p_date=parameter[str].disable_jinja_templating):
            return p_date

        with DAG(dag_id=f"current_date_dag", default_args=default_args) as dag:
            current_date(p_date="{{ ts }}")
        #### DOC END



# © Copyright Databand.ai, an IBM Company 2022

from pandas import DataFrame

from dbnd import parameter
from dbnd_spark import SparkConfig, SparkTask


class CalculateAlpha(SparkTask):
    alpha = parameter[float]

    def run(self):
        return self.alpha


class PrepareData(SparkTask):
    data = parameter[DataFrame]

    def run(self):
        return self.data


class TestDocTaskConfigurationDefaults:
    def test_doc(self):
        #### DOC START
        class CalculateBeta(CalculateAlpha):
            defaults = {
                SparkConfig.main_jar: "jar2.jar",
                # DataTask.task_env: DataTaskEnv.prod,
                CalculateAlpha.alpha: "0.5",
            }

        class PrepareData2(PrepareData):
            defaults = {
                SparkConfig.main_jar: "jar2.jar",
                # DataTask.task_env: DataTaskEnv.prod,
                CalculateAlpha.alpha: "0.5",
            }

        #### DOC END



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging

from typing import List, Tuple

import pyspark.sql as spark

from dbnd_examples_orchestration.data import demo_data_repo
from pyspark.ml import Pipeline
from pyspark.ml.classification import LogisticRegression
from pyspark.ml.feature import VectorAssembler
from pyspark.mllib.evaluation import RegressionMetrics
from pyspark.sql.functions import exp
from pyspark.sql.types import DoubleType

from dbnd import log_metric, parameter, pipeline
from dbnd.utils import data_combine, period_dates
from dbnd_spark.spark import spark_task
from targets import target
from targets.types import PathStr


logger = logging.getLogger(__name__)


def data_source(task_target_date, name):
    return target(demo_data_repo.partner_to_file(name, task_target_date))


@spark_task(result=parameter.output.csv[spark.DataFrame])
def get_and_enrich_spark(raw_data: spark.DataFrame, column_name: str):
    raw_data.show()
    data_with_new_feature = raw_data.withColumn(column_name + "_exp", exp(column_name))
    return data_with_new_feature


@spark_task(result=parameter.output.csv[spark.DataFrame])
def clean_data_spark(raw_data: spark.DataFrame):
    return raw_data.na.fill(0)


@pipeline()
def ingest_partner_a(task_target_date):
    raw_data = data_source(name="a", task_target_date=task_target_date)
    clean = clean_data_spark(raw_data=raw_data)
    return get_and_enrich_spark(raw_data=clean, column_name="1")


@pipeline
def ingest_partner_b(task_target_date):
    raw_data = data_source(name="b", task_target_date=task_target_date)
    return clean_data_spark(raw_data=raw_data)


@pipeline
def ingest_partner_c(task_target_date):
    raw_data = data_source(name="c", task_target_date=task_target_date)
    clean = clean_data_spark(raw_data=raw_data)
    return get_and_enrich_spark(raw_data=clean, column_name="10")


@pipeline
def fetch_partner_data_spark(
    task_target_date, selected_partners: List[str], period=datetime.timedelta(days=7)
) -> List[spark.DataFrame]:
    partner_data = []
    for partner in selected_partners:
        all_data = []
        for d in period_dates(task_target_date, period):
            if partner == "a":
                data = ingest_partner_a(task_target_date=d)
            elif partner == "b":
                data = ingest_partner_b(task_target_date=d)
            elif partner == "c":
                data = ingest_partner_c(task_target_date=d)
            else:
                raise Exception("Partner not found!")

            all_data.append(data)
        partner_data.append(data_combine(all_data, sort=True))
    return partner_data


@spark_task
def calculate_features(
    raw_data: List[spark.DataFrame], selected_features: List[str] = None
) -> spark.DataFrame:
    result = raw_data.pop(0)
    for d in raw_data:
        result = result.join(d, ["id"], "outer")
    if selected_features:
        result = result.select(selected_features)
    result.show()
    return result


@spark_task(result="training_set, test_set, validation_set")
def split_data_spark(
    raw_data: spark.DataFrame,
) -> Tuple[spark.DataFrame, spark.DataFrame, spark.DataFrame]:

    columns_to_remove = set(["id", "0_norm", "10_norm"])
    if columns_to_remove.issubset(list(raw_data.schema.names)):
        raw_data = raw_data.drop(columns_to_remove)

    (train, test) = raw_data.randomSplit([0.8, 0.2])
    (test, validation) = raw_data.randomSplit([0.5, 0.5])

    target_stats = raw_data.describe(["target"])

    log_metric(
        "target.mean",
        target_stats.filter(target_stats["summary"] == "mean")
        .collect()[0]
        .asDict()["target"],
    )
    log_metric(
        "target.std",
        target_stats.filter(target_stats["summary"] == "stddev")
        .collect()[0]
        .asDict()["target"],
    )

    return train, test, validation


@spark_task
def train_model_spark(
    test_set: spark.DataFrame,
    training_set: spark.DataFrame,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
    saved_model=parameter.output.folder_data.with_flag(None)[PathStr],
) -> str:

    transform = VectorAssembler(inputCols=["0", "1", "2"], outputCol="features")
    lr = LogisticRegression(
        featuresCol="features",
        labelCol="target",
        regParam=l1_ratio,
        elasticNetParam=alpha,
        family="multinomial",
        maxIter=1,
    )
    ppl = Pipeline(stages=[transform, lr])

    # Fit the pipeline to training documents.
    model = ppl.fit(training_set)

    prediction = model.transform(test_set)
    evaluation = prediction.withColumn(
        "label", prediction["target"].cast(DoubleType())
    ).select(["label", "prediction"])
    evaluation.show()
    metrics = RegressionMetrics(evaluation.rdd)

    log_metric("r2", metrics.r2)
    log_metric("alpha", alpha)

    model.write().save(str(saved_model))
    return "ok"


@pipeline(result=("model"))
def train_model_for_customer_spark(
    task_target_date,
    data: spark.DataFrame = None,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
    period=datetime.timedelta(days=1),
    selected_features: List[str] = None,
):
    if data is None:
        partners = fetch_partner_data_spark(
            task_target_date=task_target_date, period=period
        )
        data = calculate_features(
            selected_features=selected_features, raw_data=partners
        )

    training_set, test_set, validation_set = split_data_spark(raw_data=data)

    model = train_model_spark(
        test_set=test_set, training_set=training_set, alpha=alpha, l1_ratio=l1_ratio
    )

    return model.saved_model



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd_examples_orchestration.data import dbnd_examples_data_path

import dbnd

from dbnd import config


logger = logging.getLogger(__name__)


@dbnd.hookimpl
def dbnd_setup_plugin():
    # register configs
    try:
        config.set_from_config_file(dbnd_examples_data_path("examples_config.cfg"))
    except Exception:
        logger.warning(
            "Could not load examples_config.cfg! Automatic data loading for dbnd-examples is disabled!"
        )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from functools import partial

from dbnd import output, parameter, task
from dbnd_run.task.decorated_callable_task import DecoratedPythonTask


class ExperiementTask(DecoratedPythonTask):
    custom_name = parameter.value("aa")

    previous_exp = parameter.value(1)
    score_card = output.csv.data
    my_ratio = output.csv.data

    def run(self):
        # wrapping code
        score = self._invoke_func()

        self.score_card.write(str(score))
        self.my_ratio.write_pickle(self.previous_exp + 1)


experiment = partial(task, _task_type=ExperiementTask)


@experiment
def my_new_experiement(alpha: float = 0.2, previous_exp=1):
    logging.warning("My previous exp = %s", previous_exp)
    return 1, 2, alpha



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.orchestration.customizations.custom_output_factory import (
    DataSplitIntoMultipleOutputs,
)
from dbnd_examples_orchestration.orchestration.customizations.custom_output_location import (
    GenerateReportToCustomLocation,
)
from dbnd_examples_orchestration.orchestration.customizations.custom_value_type import (
    MyCustomObject,
    TaskWithCustomValue,
    TaskWithCustomValueInline,
)
from pandas import DataFrame

from dbnd_run.testing.helpers import assert_run_task
from dbnd_test_scenarios.test_common.custom_parameter_feature_store import (
    calculate_features,
    calculate_features_via_classes,
)
from dbnd_test_scenarios.test_common.custom_parameter_hdf5 import MyHdf5DataPipeline


class TestFeatureCustom(object):
    def test_data_split_into_multiple_outputs(self):
        task = assert_run_task(DataSplitIntoMultipleOutputs())
        assert "part_0" in task.splits
        assert "train_part_0" in str(task.splits["part_0"][0])

    def test_generate_report_to_different_location(self):
        task = assert_run_task(GenerateReportToCustomLocation(name="my_report"))
        assert "/my_report/" in str(task.report)

    def test_feature_store_via_func(self):
        task = assert_run_task(calculate_features.task(ratio=10))
        assert task.validation.load(str) == "OK"

    def test_feature_store_via_classes(self):
        task = assert_run_task(calculate_features_via_classes.task(ratio=10))
        assert not task.result.load(DataFrame).empty

    def test_my_data_value_type(self):
        task = assert_run_task(MyHdf5DataPipeline())
        assert not task.report.load(DataFrame).empty

    def test_custom_object_simple(self):
        t = assert_run_task(
            TaskWithCustomValue(custom_value="1", list_of_customs="['1','2']")
        )
        assert "12" == t.report.load(str)

    def test_custom_object_inline(self):
        t = assert_run_task(
            TaskWithCustomValueInline(
                custom_value=MyCustomObject("cv"),
                custom_value_with_default="cvd",
                list_of_customs=[MyCustomObject("l1"), MyCustomObject("l2")],
                set_of_customs="1,2",
                dict_of_customs="{'a':'from_dict'}",
            )
        )
        assert "cv_cvd_l1_from_dict" == t.report.load(str)

    def test_custom_config(self):
        from dbnd_examples_orchestration.orchestration.customizations.custom_paramester_user_object import (
            config_flow,
        )

        assert_run_task(config_flow.task(my_config="1"))



# © Copyright Databand.ai, an IBM Company 2022


class TestDocSparkConfiguration:
    def test_doc(self):
        """
        #### DOC START
        class PrepareData(SparkTask):
            text = parameter.data
            counters = output

            main_class = "org.predict_wine_quality.PrepareData"
            # overides value of SparkConfig object
            defaults = {SparkConfig.driver_memory: "2.5g", "spark.executor_memory": "1g"}

            def application_args(self):
                return [self.text, self.counters]
        #### DOC END
        PrepareData.dbnd_run(text="Hello!")"""



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging
import sys


logger = logging.getLogger(__name__)


def run_spark(args):
    from pyspark.sql import SparkSession
    from pyspark.sql.types import StringType, StructField, StructType

    timestamp = str(datetime.datetime.now()).replace(" ", "_")

    spark = SparkSession.builder.appName("Task_%s" % timestamp).getOrCreate()
    sc = spark.sparkContext

    veg = sc.textFile(args[1]).collect()

    mix = ["%s->%s" % (args[2], v) for v in veg]
    result = "".join(mix)

    logger.info("Dressing result %s", result)

    fields = [StructField("salad_with_dressing", StringType(), True)]
    schema = StructType(fields)

    df = spark.createDataFrame([(veg,)], schema)
    df.coalesce(1).write.csv(args[3] + timestamp)
    sc.stop()


if __name__ == "__main__":
    run_spark(sys.argv)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd import pipeline, task


logger = logging.getLogger(__name__)


@task
def calculate_alpha(base: int):
    calculated_alpha = base + 1
    logging.info("Alpha=%s", calculated_alpha)
    return calculated_alpha


@task
def calculate_ratio(base: int):
    calculated_ratio = base + 1
    logging.info("Ratio=%s", calculated_ratio)
    return calculated_ratio


@pipeline
def calculate_values_pipeline(base=1):
    alpha = calculate_alpha(base)
    ratio = calculate_ratio(base)
    return alpha, ratio


#####
# Now we can create a task that calculates base and a pipeline that wraps everything together


@task
def calculate_base(base: int = 5):
    calculated_base = base + 1
    logging.info("Base value=%s", calculated_base)
    return calculated_base


@pipeline
def calculate_values_with_base_pipeline():
    base = calculate_base()
    alpha, ratio = calculate_values_pipeline(base=base)
    return base, alpha, ratio


if __name__ == "__main__":
    result = calculate_values_with_base_pipeline.dbnd_run()
    logger.info("Result: %s", result)



# © Copyright Databand.ai, an IBM Company 2022

"""
Based on https://github.com/aws-samples/sagemaker-ml-workflow-with-apache-airflow example by Amazon.
distributed under permissive amazon licence: https://github.com/aws-samples/sagemaker-ml-workflow-with-apache-airflow/blob/master/LICENSE
"""
from __future__ import absolute_import

import io
import logging
import tempfile

from typing import Tuple

import numpy as np
import pandas as pd
import sagemaker.amazon.common as smac

from dbnd_examples_orchestration.orchestration.tool_sagemaker import sagemaker_data_repo
from scipy.sparse import lil_matrix

from dbnd import log_dataframe, log_metric, output, pipeline, task
from dbnd_aws.sagemaker import SageMakerTrainTask
from dbnd_aws.sagemaker.estimator_config import Algorithm, GenericEstimatorConfig
from targets import Target
from targets.fs import FileSystems


logger = logging.getLogger(__name__)


@task(result="training_set, test_set, validation_set")
def preprocess(
    raw_data: pd.DataFrame = sagemaker_data_repo.amazon_reviews_raw_data,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Preprocesses data based on business logic
    - Reads delimited file passed as s3_url and preprocess data by filtering
    long tail in the customer ratings data i.e. keep customers who have rated 5
    or more videos, and videos that have been rated by 9+ customers
    - Preprocessed data is then written to output

    """
    # limit dataframe to customer_id, product_id, and star_rating
    # `product_title` will be useful validating recommendations
    df = raw_data[["customer_id", "product_id", "star_rating", "product_title"]]

    # clean out the long tail because most people haven't seen most videos,
    # and people rate fewer videos than they actually watch
    customers = df["customer_id"].value_counts()
    products = df["product_id"].value_counts()

    # based on data exploration only about 5% of customers have rated 5 or
    # more videos, and only 25% of videos have been rated by 9+ customers
    customers = customers[customers >= 5]
    products = products[products >= 10]
    log_dataframe("Original data shape", df)

    reduced_df = df.merge(pd.DataFrame({"customer_id": customers.index})).merge(
        pd.DataFrame({"product_id": products.index})
    )

    log_dataframe("Shape after removing long tail", reduced_df)

    reduced_df = reduced_df.drop_duplicates(["customer_id", "product_id"])
    log_dataframe("Shape after removing duplicates", reduced_df)

    # recreate customer and product lists since there are customers with
    # more than 5 reviews, but all of their reviews are on products with
    # less than 5 reviews (and vice versa)
    customers = reduced_df["customer_id"].value_counts()
    products = reduced_df["product_id"].value_counts()

    # sequentially index each user and item to hold the sparse format where
    # the indices indicate the row and column in our ratings matrix
    customer_index = pd.DataFrame(
        {"customer_id": customers.index, "customer": np.arange(customers.shape[0])}
    )
    product_index = pd.DataFrame(
        {"product_id": products.index, "product": np.arange(products.shape[0])}
    )
    reduced_df = reduced_df.merge(customer_index).merge(product_index)

    nb_customer = reduced_df["customer"].max() + 1
    nb_products = reduced_df["product"].max() + 1
    feature_dim = nb_customer + nb_products

    log_metric(
        "features(customer,product,total)", (nb_customer, nb_products, feature_dim)
    )
    # print(nb_customer, nb_products, feature_dim)

    product_df = reduced_df[["customer", "product", "star_rating"]]

    # split into train, validation and test data sets
    train_df, validate_df, test_df = np.split(
        product_df.sample(frac=1),
        [int(0.6 * len(product_df)), int(0.8 * len(product_df))],
    )

    log_metric("# of rows train", train_df.shape[0])
    log_metric("# of rows test", test_df.shape[0])
    log_metric("# of rows validation", validate_df.shape[0])

    # select columns required for training the model
    # excluding columns "customer_id", "product_id", "product_title" to
    # keep files small
    cols = ["customer", "product", "star_rating"]
    train_df = train_df[cols]
    validate_df = validate_df[cols]
    test_df = test_df[cols]

    return train_df, test_df, validate_df


def convert_sparse_matrix(df, nb_rows, nb_customer, nb_products):
    # dataframe to array
    df_val = df.values

    # determine feature size
    nb_cols = nb_customer + nb_products
    log_metric("# of rows", nb_rows)
    log_metric("# of cols", nb_cols)

    # extract customers and ratings
    df_X = df_val[:, 0:2]
    # Features are one-hot encoded in a sparse matrix
    X = lil_matrix((nb_rows, nb_cols)).astype("float32")
    df_X[:, 1] = nb_customer + df_X[:, 1]
    coords = df_X[:, 0:2]
    X[np.arange(nb_rows), coords[:, 0]] = 1
    X[np.arange(nb_rows), coords[:, 1]] = 1

    # create label with ratings
    Y = df_val[:, 2].astype("float32")

    # validate size and shape

    logger.info("Shape of X: %s ", str(X.shape))
    logger.info("Shape of Y: %s", str(Y.shape))
    assert X.shape == (nb_rows, nb_cols)
    assert Y.shape == (nb_rows,)

    return X, Y


def save_as_protobuf(X, Y, target):
    """Converts features and predictions matrices to recordio protobuf
    Args:
        X:
          2D numpy matrix with features
        Y:
          1D numpy matrix with predictions
        target:
          protobuf file name to be staged
    """
    buf = io.BytesIO()
    smac.write_spmatrix_to_sparse_tensor(buf, X, Y)

    f, name = tempfile.mkstemp()
    with open(f, "wb") as fd:
        fd.write(buf.getvalue())
        fd.seek(0)
        if target.fs.name == FileSystems.s3:
            target.fs.put(name, str(target))
        else:
            target.move_from(from_path=name)


def chunk(x, batch_size):
    """split array into chunks of batch_size"""
    chunk_range = range(0, x.shape[0], batch_size)
    chunks = [x[p : p + batch_size] for p in chunk_range]
    return chunks


@task
def prepare(
    train: pd.DataFrame,
    test: pd.DataFrame,
    validation: pd.DataFrame,
    train_out=output[Target],
    test_out=output.folder_data.with_flag(None)[Target],
    validation_out=output[Target],
):
    """Prepare data for training with Sagemaker algorithms

    - Read preprocessed data and converts to ProtoBuf format to prepare for
      training with Sagemaker algorithms

    Args:
        test - dataframe with test data
        train  - dataframe with train data
        validation - dataframe with validation data
        train_out - output path for train data in protobuf format
        test_out - output path for test data in protobuf format
        validation_out - output path for validation data in protobuf format
    Returns:
        s3 url with key to the prepared data
    """
    all_df = pd.concat([train, validation, test])
    nb_customer = np.unique(all_df["customer"].values).shape[0]
    nb_products = np.unique(all_df["product"].values).shape[0]
    feature_dim = nb_customer + nb_products
    log_metric(
        "customers x products x feature_dim", (nb_customer, nb_products, feature_dim)
    )

    train_X, train_Y = convert_sparse_matrix(
        train, train.shape[0], nb_customer, nb_products
    )
    validate_X, validate_Y = convert_sparse_matrix(
        validation, validation.shape[0], nb_customer, nb_products
    )
    test_X, test_Y = convert_sparse_matrix(
        test, test.shape[0], nb_customer, nb_products
    )

    save_as_protobuf(train_X, train_Y, train_out)
    save_as_protobuf(validate_X, validate_Y, validation_out)

    test_x_chunks = chunk(test_X, 10000)
    test_y_chunks = chunk(test_Y, 10000)
    N = len(test_x_chunks)
    for i in range(N):
        save_as_protobuf(test_x_chunks[i], test_y_chunks[i], test_out.partition())
    return "OK"


@pipeline
def train_review_model(
    raw_data: pd.DataFrame = sagemaker_data_repo.amazon_reviews_raw_data,
):
    train, test, validate = preprocess(raw_data=raw_data)
    prepared = prepare(train=train, test=test, validation=validate)

    hyperparameters = {
        "feature_dim": "178729",
        "epochs": "10",
        "mini_batch_size": "200",
        "num_factors": "64",
        "predictor_type": "regressor",
    }

    train = SageMakerTrainTask(
        train=prepared.train_out,
        estimator_config=GenericEstimatorConfig(
            algorithm=Algorithm.FACTORIZATION_MACHINES, hyperparameters=hyperparameters
        ),
    )

    return prepared, train.output_path



# © Copyright Databand.ai, an IBM Company 2022

import pytest

from dbnd import config, relative_path
from dbnd_run.testing.helpers import assert_run_task


def _data_for_spark_path(*path):
    from dbnd_examples_orchestration.orchestration.dbnd_spark import (
        read_from_multiple_sources,
    )

    return relative_path(
        read_from_multiple_sources.__file__, "data_for_spark_examples", *path
    )


@pytest.mark.spark
class TestPysparkTasks(object):
    def test_spark_inline_same_context(self):
        from dbnd_examples_orchestration.orchestration.dbnd_spark.word_count import (
            word_count_inline,
        )
        from pyspark.sql import SparkSession

        from dbnd_spark.local.local_spark_config import SparkLocalEngineConfig

        with SparkSession.builder.getOrCreate() as sc:
            with config({SparkLocalEngineConfig.enable_spark_context_inplace: True}):
                task_instance = word_count_inline.t(text=__file__)
                assert_run_task(task_instance)

    def test_read_from_multiple_sources(self):
        from dbnd_examples_orchestration.orchestration.dbnd_spark.read_from_multiple_sources import (
            data_source_complicated_pipeline,
        )
        from pyspark.sql import SparkSession

        from dbnd_spark.local.local_spark_config import SparkLocalEngineConfig

        with SparkSession.builder.getOrCreate() as sc:
            with config({SparkLocalEngineConfig.enable_spark_context_inplace: True}):
                task_instance = data_source_complicated_pipeline.t(
                    root_path=_data_for_spark_path("read_from_multiple_sources"),
                    extra_data=_data_for_spark_path(
                        "read_from_multiple_sources1", "configID=1", "1.tsv"
                    ),
                )
                assert_run_task(task_instance)



# © Copyright Databand.ai, an IBM Company 2022

import gzip
import struct

import mxnet as mx
import numpy as np

from mxnet import nd
from mxnet.gluon import utils


def read_minst(data_file, label_file, gluon_format=False):
    with gzip.open(label_file, "rb") as fin:
        struct.unpack(">II", fin.read(8))
        label = np.frombuffer(fin.read(), dtype=np.uint8).astype(np.int32)
    with gzip.open(data_file, "rb") as fin:
        struct.unpack(">IIII", fin.read(16))
        data = np.frombuffer(fin.read(), dtype=np.uint8)
        if gluon_format:
            data = data.reshape(len(label), 28, 28, 1)
            data = nd.array(data, dtype=data.dtype)
        else:
            data = data.reshape(len(label), 1, 28, 28).astype(np.float32) / 255
    return data, label


def mnist_to_ndarray_iterator(data_file, label_file, batch_size, shuffle=False):
    data, label = read_minst(data_file, label_file, gluon_format=False)
    return mx.io.NDArrayIter(data, label, batch_size, shuffle)


class FashionMNIST(object):
    namespace = "gluon/dataset/fashion-mnist"

    train_data = (
        utils._get_repo_file_url(namespace, "train-images-idx3-ubyte.gz"),
        "0cf37b0d40ed5169c6b3aba31069a9770ac9043d",  # pragma: allowlist secret
    )
    train_label = (
        utils._get_repo_file_url(namespace, "train-labels-idx1-ubyte.gz"),
        "236021d52f1e40852b06a4c3008d8de8aef1e40b",  # pragma: allowlist secret
    )
    test_data = (
        utils._get_repo_file_url(namespace, "t10k-images-idx3-ubyte.gz"),
        "626ed6a7c06dd17c0eec72fa3be1740f146a2863",  # pragma: allowlist secret
    )
    test_label = (
        utils._get_repo_file_url(namespace, "t10k-labels-idx1-ubyte.gz"),
        "17f9ab60e7257a1620f4ad76bbbaf857c3920701",  # pragma: allowlist secret
    )


class MNIST(object):
    namespace = "gluon/dataset/mnist"

    train_data = (
        utils._get_repo_file_url(namespace, "train-images-idx3-ubyte.gz"),
        "6c95f4b05d2bf285e1bfb0e7960c31bd3b3f8a7d",  # pragma: allowlist secret
    )
    train_label = (
        utils._get_repo_file_url(namespace, "train-labels-idx1-ubyte.gz"),
        "2a80914081dc54586dbdf242f9805a6b8d2a15fc",  # pragma: allowlist secret
    )
    test_data = (
        utils._get_repo_file_url(namespace, "t10k-images-idx3-ubyte.gz"),
        "c3a25af1f52dad7f726cce8cacb138654b760d48",  # pragma: allowlist secret
    )
    test_label = (
        utils._get_repo_file_url(namespace, "t10k-labels-idx1-ubyte.gz"),
        "763e7fa3757d93b0cdec073cef058b2004252c17",  # pragma: allowlist secret
    )


fashion_data = FashionMNIST()
digit_data = MNIST()



# © Copyright Databand.ai, an IBM Company 2022

import logging

import pyspark.sql as spark
import six

from databand import output, parameter
from dbnd import config, dbnd_context
from dbnd._core.task_build.task_metaclass import TaskMetaclass
from dbnd_spark import SparkConfig
from dbnd_spark.spark import PySparkInlineTask, spark_task
from targets.target_config import FileFormat


logger = logging.getLogger(__name__)


@spark_task(result=output.save_options(FileFormat.csv, header=True)[spark.DataFrame])
def custom_load_save_options(
    data=parameter.load_options(FileFormat.csv, header=False, sep="\t")[spark.DataFrame]
):
    print(data.show())
    return data


class AdvancedSparkConfigExample(PySparkInlineTask):
    text = parameter.csv[spark.DataFrame]
    counters = output.txt.data
    my_p = parameter.value(3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Every parameter of the task can be changed including spark config.
        # However, you can not use "configuration" (SparkConfig.driver_memory)
        # for that, as all config objects are calculated and created already.

        # we can change some spark config param based on other param
        if self.env == "local":
            self.spark_config.driver_memory = "2G"
        else:
            self.spark_config.driver_memory = "4G"

        self.my_p = 10

    def run(self):
        """simple map function, it will run inside spark context (after submission by dbnd)"""
        from operator import add

        lines = self.text.rdd.map(lambda r: r[0])
        counts = (
            lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
        )
        counts.saveAsTextFile(str(self.counters))


#######################
## very advanced usage
## add extra config to spark tasks based on task class and env


class AdvancedConfigTaskMetaclass(TaskMetaclass):
    """
    The Metaclass of :py:class:`Task`.
    """

    def __call__(cls, *args, **kwargs):
        """
        Extension of Task Metaclass, we want to add our own configuration based on some values
        so the moment task is created, that config is used.
        """

        config_values = cls.get_custom_config()

        # create new config layer, so when we are out of this process -> config is back to the previous value
        with config(
            config_values=config_values,
            source=cls.task_definition.task_passport.format_source_name(
                ".get_custom_config"
            ),
        ):
            return super(AdvancedConfigTaskMetaclass, cls).__call__(*args, **kwargs)


@six.add_metaclass(AdvancedConfigTaskMetaclass)
class PySparkInlineWithAdvancedConfig(PySparkInlineTask):
    """you can apply AdvancedConfigTaskMetaclass to specific task or to some `base` class"""

    @classmethod
    def get_custom_config(cls):
        current_env = dbnd_context().env
        logger.info("Creating task with spark config for %s", current_env)
        if current_env.name == "local":
            return {SparkConfig.num_executors: 3}
        return {SparkConfig.num_executors: 5}


class AdvancedSparkConfigAtMetaclassExample(PySparkInlineWithAdvancedConfig):
    text = parameter.csv[spark.DataFrame]
    counters = output.txt.data
    my_p = parameter.value(3)

    def run(self):
        """simple map function, it will run inside spark context (after submission by dbnd)"""
        from operator import add

        lines = self.text.rdd.map(lambda r: r[0])
        counts = (
            lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
        )
        counts.saveAsTextFile(str(self.counters))



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd_examples_orchestration.orchestration.tutorial_syntax import T03_task_outputs


logger = logging.getLogger(__name__)


class TestTutorialSyntax(object):
    def test_t23_task_with_multiple_outputs_v1(self):
        t = T03_task_outputs.f_returns_two_dataframes_v1.dbnd_run(1).task
        assert t.features

    def test_t23_task_with_multiple_outputs_v2(self):
        t = T03_task_outputs.f_returns_two_dataframes_v2.dbnd_run(2).task
        assert t.features

    def test_t23_task_with_multiple_outputs_no_hint(self):
        t = T03_task_outputs.f_returns_two_dataframes_no_hint.dbnd_run(2).task
        assert t.result_1

    def test_t23_task_with_multiple_outputs_named_tuple_v1(self):
        t = T03_task_outputs.f_returns_two_dataframes_named_tuple_v1.dbnd_run(2).task
        assert t.features

    def test_t23_task_with_multiple_outputs_named_tuple_v2(self):
        t = T03_task_outputs.f_returns_two_dataframes_named_tuple_v2.dbnd_run(2).task
        assert t.features



# © Copyright Databand.ai, an IBM Company 2022

from datetime import datetime

from dbnd_examples_orchestration.data import data_repo
from dbnd_examples_orchestration.tests.documentation.orchestration.test_quick_start_wine_quality import (
    prepare_data,
    train_model,
    validate_model,
)
from pandas import DataFrame
from sklearn.linear_model import ElasticNet

from dbnd import pipeline


class TestDocAccessingRunResults:
    def test_doc(self):
        #### DOC START
        @pipeline(result=("model", "validation"))
        def predict_wine_quality(
            raw_data: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
        ):
            training_set, validation_set = prepare_data(raw_data=raw_data)
            model = train_model(
                training_set=training_set, alpha=alpha, l1_ratio=l1_ratio
            )

            validation = validate_model(model=model, validation_dataset=validation_set)

            return model, validation

        run = predict_wine_quality.dbnd_run(
            raw_data=data_repo.wines, task_version=datetime.now()
        )

        model = run.load_from_result("model", value_type=ElasticNet)
        validation = run.load_from_result("validation")
        #### DOC END
        assert model and validation



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Dict

import numpy
import pandas as pd

from dbnd import pipeline, task


logger = logging.getLogger(__name__)


@task
def f_pandas_df(i: int) -> pd.DataFrame:
    return pd.DataFrame(
        data=list(zip(["Bob", "Jessica"], [i, i])), columns=["Names", "Births"]
    )


@task
def f_df_dict(df_dict):
    # type: (Dict[str, pd.DataFrame]) -> pd.DataFrame
    values = df_dict.values()
    return pd.concat(values)


@pipeline
def f_pipe_df_dict():
    df_dict = {}
    for i in range(5):
        df_dict[str(i)] = f_pandas_df(i)
    df_sum = f_df_dict(df_dict)

    return df_sum


@task
def f_pandas_numpy(v_df: pd.DataFrame, v_series: pd.Series, v_np: numpy.ndarray) -> str:
    assert isinstance(v_df, pd.DataFrame)
    assert isinstance(v_series, pd.Series)
    assert isinstance(v_np, numpy.ndarray)

    return "df:%s series:%s np:%s" % (v_df.shape, v_series.shape, v_np.shape)


@task(result=("df", "series", "np"))
def f_create_pandas_numpy() -> (pd.DataFrame, pd.Series, numpy.ndarray):
    s = pd.Series([1, 3, 5, 8])
    df = pd.DataFrame(
        data=list(zip(["Bob", "Jessica"], [968, 155])), columns=["Names", "Births"]
    )

    return df, s, df.values


@pipeline
def f_run_pandas_numpy():
    simple = f_create_pandas_numpy()
    pandas_numpy_str = f_pandas_numpy(*simple)

    return (pandas_numpy_str,)


@task
def f_assert(pandas_numpy_str):
    assert "df:(2, 2) series:(4,) np:(2, 2)" == pandas_numpy_str

    return "OK"


@pipeline
def f_test_pandas_numpy_flow():
    all_simple = f_run_pandas_numpy()
    return f_assert(*all_simple)


if __name__ == "__main__":
    # we can just run the function!
    f_test_pandas_numpy_flow()
    f_pipe_df_dict.dbnd_run()



# © Copyright Databand.ai, an IBM Company 2022

#### DOC START

"""[kubernetes_engine] # make sure it's your engine name

secrets = [
   { "type":"env", "target": "AIRFLOW__CORE__SQL_ALCHEMY_CONN", "secret" : "databand-secret", "key": "airflow_sql_alchemy_conn"},
   { "type":"env", "target": "AIRFLOW__CORE__FERNET_KEY", "secret" : "databand-secret", "key": "airflow_fernet_key"},
   { "type":"env", "target": "DBND__CORE__DATABAND_URL", "secret" : "databand-secret", "key": "databand_url"}
   ]

"""

#### DOC END



# © Copyright Databand.ai, an IBM Company 2022

# This example requires pandas, numpy, sklearn, scipy
# Inspired by an MLFlow tutorial:
#   https://github.com/databricks/mlflow/blob/master/example/tutorial/train.py
import datetime
import itertools
import logging
import sys

from typing import Tuple

import numpy as np
import pandas as pd

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import train_test_split

from dbnd import (
    dbnd_config,
    dbnd_handle_errors,
    log_dataframe,
    log_metric,
    output,
    parameter,
    pipeline,
    task,
)
from dbnd.utils import data_combine, period_dates
from targets import target
from targets.types import PathStr


logger = logging.getLogger(__name__)


# dbnd run dbnd_examples_orchestration.orchestration.examples.wine_quality.predict_wine_quality --task-version now
# dbnd run dbnd_examples_orchestration.orchestration.examples.wine_quality.predict_wine_quality_parameter_search --task-version now


def calculate_metrics(actual, pred):
    rmse = np.sqrt(mean_squared_error(actual, pred))
    mae = mean_absolute_error(actual, pred)
    r2 = r2_score(actual, pred)
    return rmse, mae, r2


@task(result="training_set, test_set, validation_set")
def prepare_data(raw_data: DataFrame) -> Tuple[DataFrame, DataFrame, DataFrame]:
    """Split data into train, test and validation"""
    train_df, test_df = train_test_split(raw_data)
    test_df, validation_df = train_test_split(test_df, test_size=0.5)

    sys.stderr.write("Running Prepare Data! You'll see this message in task log \n")
    print("..and this one..\n")
    logger.info("..and this one for sure!")

    log_dataframe("raw", raw_data, with_histograms=True)

    return train_df, test_df, validation_df


@task
def calculate_alpha(alpha: float = 0.5) -> float:
    """Calculates alpha for train_model"""
    alpha += 0.1
    return alpha


@task(training_set=parameter[DataFrame](log_histograms=True, log_stats=True))
def train_model(
    test_set: DataFrame,
    training_set: DataFrame,
    alpha: float = 0.5,
    l1_ratio: float = 0.5,
) -> ElasticNet:
    """Train wine prediction model"""
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
    lr.fit(training_set.drop(["quality"], 1), training_set[["quality"]])
    prediction = lr.predict(test_set.drop(["quality"], 1))

    (rmse, mae, r2) = calculate_metrics(test_set[["quality"]], prediction)

    log_metric("alpha", alpha)
    log_metric("rmse", rmse)
    log_metric("mae", rmse)
    log_metric("r2", r2)

    logging.info(
        "Elasticnet model (alpha=%f, l1_ratio=%f): rmse = %f, mae = %f, r2 = %f",
        alpha,
        l1_ratio,
        rmse,
        mae,
        r2,
    )
    return lr


@task
def validate_model(model: ElasticNet, validation_dataset: DataFrame) -> str:
    """Calculates metrics of wine prediction model"""
    log_dataframe("validation", validation_dataset)
    # support for py3 parqeut
    validation_dataset = validation_dataset.rename(str, axis="columns")
    validation_x = validation_dataset.drop(["quality"], 1)
    validation_y = validation_dataset[["quality"]]

    prediction = model.predict(validation_x)
    (rmse, mae, r2) = calculate_metrics(validation_y, prediction)

    log_metric("rmse", rmse)
    log_metric("mae", rmse)
    log_metric("r2", r2)

    return "%s,%s,%s" % (rmse, mae, r2)


@pipeline(result=("model", "validation"))
def predict_wine_quality(
    data: DataFrame = None,
    alpha: float = 0.5,
    l1_ratio: float = 0.5,
    good_alpha: bool = False,
):
    """Entry point for wine quality prediction"""
    if data is None:
        data = fetch_data()
    training_set, test_set, validation_set = prepare_data(raw_data=data)

    if good_alpha:
        alpha = calculate_alpha(alpha)

    model = train_model(
        test_set=test_set, training_set=training_set, alpha=alpha, l1_ratio=l1_ratio
    )
    validation = validate_model(model=model, validation_dataset=validation_set)
    return model, validation


@pipeline
def predict_wine_quality_parameter_search(
    alpha_step: float = 0.3, l1_ratio_step: float = 0.4
):
    result = {}
    variants = list(
        itertools.product(np.arange(0, 1, alpha_step), np.arange(0, 1, l1_ratio_step))
    )
    logger.info("All Variants: %s", variants)
    for alpha_value, l1_ratio in variants:
        exp_name = "Predict_%f_l1_ratio_%f" % (alpha_value, l1_ratio)
        model, validation = predict_wine_quality(
            alpha=alpha_value, l1_ratio=l1_ratio, task_name=exp_name
        )

        result[exp_name] = (model, validation)
    return result


# DATA FETCHING
@pipeline
def wine_quality_day(
    task_target_date: datetime.date, root_location: PathStr = data_repo.wines_per_date
) -> pd.DataFrame:
    return target(root_location, task_target_date.strftime("%Y-%m-%d"), "wine.csv")


@task(result=output.prod_immutable[DataFrame](log_histograms=True))
def fetch_wine_quality(
    task_target_date: datetime.date, data: pd.DataFrame = data_repo.wines_full
) -> pd.DataFrame:
    # very simple implementation that just sampe the data with seed = target date
    return DataFrame.sample(data, frac=0.2, random_state=task_target_date.day)


@pipeline
def fetch_data(task_target_date, period=datetime.timedelta(days=7)):
    all_data = []
    for d in period_dates(task_target_date, period):
        data = fetch_wine_quality(task_target_date=d)
        all_data.append(data)

    return data_combine(all_data, sort=True)


def main():
    # NATIVE EXECUTION
    print(predict_wine_quality(pd.read_csv(data_repo.wines)))


@dbnd_handle_errors()
def main_dbnd():
    # DBND EXECUTION
    with dbnd_config({fetch_data.task.period: "2d"}):
        predict_wine_quality.dbnd_run()


if __name__ == "__main__":
    main()



# © Copyright Databand.ai, an IBM Company 2022

#### DOC START

from dbnd import current_task, task


@task
def calculate_alpha(alpha: float = 0.5):
    return current_task().task_env.name  # The environment of the task
    # See EnvConfig object for all properties


#### DOC END


class TestDocFAQ:
    def test_doc(self):
        calculate_alpha.dbnd_run()



# © Copyright Databand.ai, an IBM Company 2022

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by databand.ai to for more advance example.

from __future__ import absolute_import

import logging
import re

import apache_beam as beam

from apache_beam.io import ReadFromText, WriteToText
from apache_beam.metrics import Metrics
from apache_beam.metrics.metric import MetricsFilter
from past.builtins import unicode

from dbnd import dbnd_main
from dbnd_gcp.apache_beam.apache_beam_task import beam_task
from dbnd_gcp.apache_beam.parameters import beam_output
from targets import target
from targets.types import PathStr


class WordExtractingDoFn(beam.DoFn):
    """Parse each line of input text into words."""

    def __init__(self):
        self.words_counter = Metrics.counter(self.__class__, "words")
        self.word_lengths_counter = Metrics.counter(self.__class__, "word_lengths")
        self.word_lengths_dist = Metrics.distribution(self.__class__, "word_len_dist")
        self.empty_line_counter = Metrics.counter(self.__class__, "empty_lines")

    def process(self, element):
        """Returns an iterator over the words of this element.

        The element is a line of text.  If the line is blank, note that, too.

        Args:
          element: the element being processed

        Returns:
          The processed element.
        """
        text_line = element.strip()
        if not text_line:
            self.empty_line_counter.inc(1)
        words = re.findall(r"[\w\']+", text_line, re.UNICODE)
        for w in words:
            self.words_counter.inc()
            self.word_lengths_counter.inc(len(w))
            self.word_lengths_dist.update(len(w))
        return words


@beam_task
def word_count(dataflow_pipeline, text_input, text_output=beam_output):
    # type:(Any, PathStr, PathStr) -> None
    # Read the text file[pattern] into a PCollection.
    lines = dataflow_pipeline | "read" >> ReadFromText(text_input)

    # Count the occurrences of each word.
    def count_ones(word_ones):
        (word, ones) = word_ones
        return (word, sum(ones))

    counts = (
        lines
        | "split" >> (beam.ParDo(WordExtractingDoFn()).with_output_types(unicode))
        | "pair_with_one" >> beam.Map(lambda x: (x, 1))
        | "group" >> beam.GroupByKey()
        | "count" >> beam.Map(count_ones)
    )

    # Format the counts into a PCollection of strings.
    def format_result(word_count):
        (word, count) = word_count
        return "%s: %d" % (word, count)

    output = counts | "format" >> beam.Map(format_result)

    # Write the output using a "Write" transform that has side effects.
    # pylint: disable=expression-not-assigned

    output | "write" >> WriteToText(text_output)

    target(text_output).mkdir()


@beam_task(dataflow_wait_until_finish=False)
def word_count_with_metrics(text_input, dataflow_pipeline, text_output=beam_output):
    word_count(
        text_input=text_input,
        text_output=text_output,
        dataflow_pipeline=dataflow_pipeline,
    )
    result = dataflow_pipeline.run()
    result.wait_until_finish()

    # Do not query metrics when creating a template which doesn't run
    if (
        not hasattr(result, "has_job") or result.has_job  # direct runner
    ):  # not just a template creation
        empty_lines_filter = MetricsFilter().with_name("empty_lines")
        query_result = result.metrics().query(empty_lines_filter)
        if query_result["counters"]:
            empty_lines_counter = query_result["counters"][0]
            logging.info("number of empty lines: %d", empty_lines_counter.result)

        word_lengths_filter = MetricsFilter().with_name("word_len_dist")
        query_result = result.metrics().query(word_lengths_filter)
        if query_result["distributions"]:
            word_lengths_dist = query_result["distributions"][0]
            logging.info("average word length: %d", word_lengths_dist.result.mean)


if __name__ == "__main__":
    dbnd_main()
    # word_count.dbnd_run(
    #     # text_input="~/data/kinglear.txt",
    #     text_input="gs://dataflow-samples/shakespeare/kinglear.txt",
    #     task_env="gcp",
    #     task_version="now",
    # )



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_gcp import BigQuery


class FetchDataFromBQ(BigQuery):
    """Fetch wine quality data from Google's BigQuery"""

    query = "SELECT * FROM `wine_ds.wine_table`"



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Dict, List

import matplotlib
import pandas as pd

from matplotlib.figure import Figure
from pandas import DataFrame, np

from dbnd import band, log_dataframe, output, task
from dbnd.utils import data_combine
from dbnd_run.tasks.basics.publish import publish_results
from targets import Target


matplotlib.use("Agg")
logger = logging.getLogger(__name__)


@task
def dump_table_from_db(table, db_name="master"):
    # type: (str, str) -> List[str]
    # output: raw_id(table+id), partner, value
    return ["raw_id,partner,value\n"] + [
        "%s_%s,%s,%s\n" % (table, idx, idx % 3, idx * 2) for idx in range(10)
    ]


@band
def dump_db(tables, db_name="master"):
    # type: (List[str], str) -> Dict[str, pd.DataFrame]
    result = {
        t: dump_table_from_db(table=t, db_name=db_name, task_name="table_%s" % t)
        for t in tables
    }
    return result


@task
def filter_partner(data, partner):
    # type: (DataFrame, int) -> DataFrame
    partner_data = data[data["partner"] == partner]
    log_dataframe("partner_%s" % partner, partner_data)
    return partner_data


@task(result=output.folder[Dict[str, DataFrame]])
def separate_data_by_type(data):
    # type: (DataFrame) -> Dict[str, DataFrame]

    return {"a": data.head(1), "b": data.head(2), "c": data.head(1)}


@task
def build_report(partners_files, partners_df):
    # type: (Dict[int,Target], DataFrame) -> str
    logger.info("partners files: %s", partners_files)
    logger.info("partners data: %s", partners_df)
    return "OK"


@task
def build_graphs(partner_info):
    # type: (DataFrame) -> Figure
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax1 = fig.add_subplot(2, 2, 1)
    ax1.hist(np.random.randn(100), bins=20, alpha=0.3)
    ax2 = fig.add_subplot(2, 2, 2)
    ax2.scatter(np.arange(30), np.arange(30) + 3 * np.random.randn(30))
    fig.add_subplot(2, 2, 3)
    return fig


@band
def partners_report(partners):
    # type: (List[int]) -> ...
    table = "aa"

    dump = dump_db(tables=[table])
    partners_info = {
        partner: filter_partner(dump[table], partner) for partner in partners
    }
    by_type = separate_data_by_type(dump[table])

    report = build_report(
        partners_files=partners_info, partners_df=data_combine(partners_info.values())
    )

    graphs = {partner: build_graphs(pi) for partner, pi in partners_info.items()}

    published = publish_results(graphs).published

    return report, by_type, published



# © Copyright Databand.ai, an IBM Company 2022

from typing import List

from dbnd_examples_orchestration.orchestration.examples.top_artists import (
    top_artists_big_report,
    top_artists_report,
)

from dbnd_run.testing.helpers import assert_run_task


class TestTopArtistExamples(object):
    def test_top_artists_report(self):
        task = assert_run_task(top_artists_report.task())
        actual = task.result.load(List[str])
        assert len(actual) == 10

    def test_top_artists_big_report(self):
        task = assert_run_task(top_artists_big_report.task())
        actual = task.result.load(List[str])
        assert len(actual) == 15



# © Copyright Databand.ai, an IBM Company 2022

import logging

from functools import partial

from dbnd import current_task, parameter, task
from dbnd._core.task.decorated_callable_task import DecoratedPythonTask
from targets.values import ValueType, register_value_type


# INFRA OBJECTS
class MyConfig(object):
    def __init__(self, config_id, config_ratio):
        self.config_id = config_id
        self.config_ratio = config_ratio

    def __str__(self):
        return "CONFIG: %s, %s" % (self.config_id, self.config_ratio)


class MyConfigTypeHandler(ValueType):
    type = MyConfig

    def parse_from_str(self, x):
        logging.info("Loading config %s from DB..", x)
        return MyConfig(config_id=x, config_ratio=0.7)

    def to_str(self, x):
        return str(x)

    def to_signature(self, x):
        return x.config_id


register_value_type(MyConfigTypeHandler())


class MyProjectTask(DecoratedPythonTask):
    config = parameter[MyConfig]

    def should_rerun(self):
        return self.config.config_ratio == 0.2

    def on_kill(self):
        pass


my_project_task = partial(task, _task_type=MyProjectTask)


# USAGE EXAMPLE
@my_project_task
def my_new_experiement(alpha: float = 0.2, config=parameter[MyConfig]):
    logging.warning("My current config= %s", config)
    return 1, 2, alpha


@my_project_task
class MyAnotherExperiement(object):
    def __init__(self, config, ratio=0.2):
        self.ratio = ratio
        logging.info("Creating MyAnotherExperiement with config: %s", config)
        self.config_id = config.config_id
        super(MyAnotherExperiement, self).__init__()

    def run(self):
        # current config
        t = current_task()

        logging.info("Config : %s", t.config)
        logging.info("Ratio : %s", self.ratio)



# © Copyright Databand.ai, an IBM Company 2022

import logging

import six

from dbnd import (
    ParameterDefinition,
    PipelineTask,
    PythonTask,
    output,
    parameter,
    pipeline,
)


def _custom_outputs_factory(
    task, task_output
):  # type: (PythonTask, ParameterDefinition) -> dict
    result = {}
    target = task_output.build_target(task)
    for i in range(task.parts):
        name = "part_%s" % i
        result[name] = (
            target.partition(name="train_%s" % name),
            target.partition(name="test_%s" % name),
        )

    return result


class DataSplitIntoMultipleOutputs(PythonTask):
    parts = parameter.value(3)
    splits = output.csv.folder(output_factory=_custom_outputs_factory)

    def run(self):
        for key, split in self.splits.items():
            train, test = split
            train.write(key)
            test.write(key)


def _get_all_splits(task, task_output):  # type: (Task, Any) -> Dict[str,Target]
    result = {}
    target = task_output.build_target(task)
    for i in range(task.splits_count):
        result[str(i)] = target.partition(name="split_%s" % i)
    return result


class DataSplit(PythonTask):
    splits_count = parameter.default(2)
    splits = output.csv.folder(output_factory=_get_all_splits)

    def run(self):
        for key, split in six.iteritems(self.splits):
            logging.info("writing split")
            split.write("split_%s" % key)


# we can implement using classes
# we will treat output as a Dict of str:result (result is Target, deferred data result)
# while running band it still doesnt' exist, as we are only building a graph (we are before exectuion)
class DataSplitBand(PipelineTask):
    split_selector = parameter[str]
    selected_split = output

    def band(self):
        logging.info(
            "All logs will be printed on task creation - before the real execution"
        )
        splits = DataSplit().splits
        for key, split in six.iteritems(splits):
            logging.info("split %s %s", key, split)

        self.selected_split = splits[self.split_selector]


# here we wrap the pipeline -- if we want to do that
class DataSplitUsage(PipelineTask):
    selected_split = output

    def band(self):
        self.selected_split = DataSplitBand(split_selector="0").selected_split


# and here the regular dbnd pipeline implementation,  notice, we are selecting part "1" this time
@pipeline
def data_split_selector():
    return DataSplit().splits["1"]
